"""Tools to generate music"""

from tonal.chords import chords_to_wav, chords_to_midi
from tonal.util import DFLT_SOUNDFONT
from tonal.converters import convert, midi_to_wav
from tonal.notes import scale_midi_notes



"""Utils for sonification."""

# TODO: The big module (and even package) todo is to refactor everything to do with
#   the hierarchy of iterables (Score=I[Part]), Part=I[Measure], Measure=I[Note])
#   to be consistent.

import os
from typing import Callable
from importlib.resources import files
from functools import partial
from config2py import (
    process_path,
    simple_config_getter,
    get_app_data_folder,
)
from config2py.errors import ConfigNotFound

from dol import written_bytes


# from typing import Sequence, Callable, Tuple
# from music21.note import Note
# Chord = str
# Notes = Sequence[Note]
# ChordTimed = Tuple[Chord, Notes]
# ChordSequence = Sequence[ChordTimed]
# ChordDefinitions = Callable[[Chord], Notes]
# ChordRenderer = Callable[[Notes, any, int], None]

pkg_name = "tonal"

app_data_dir = process_path(get_app_data_folder("tonal"))
soundfonts_dir = process_path(
    os.path.join(app_data_dir, "soundfonts"), ensure_dir_exists=True
)

data_files = files(pkg_name) / "data"

get_config = simple_config_getter(pkg_name)


DFLT_OUTPUT_NAME = "audio_output"
DFLT_MIDI_OUTPUT = f"{DFLT_OUTPUT_NAME}.mid"
DFLT_WAV_OUTPUT = f"{DFLT_OUTPUT_NAME}.wav"

# TODO: The DFLT_SOUNDFONT is horribly unclean. Revise. config2py should have an easy default mechanism
try:
    DFLT_SOUNDFONT = process_path(
        get_config("TONAL_DFLT_SOUNDFONT_PATH"),
    )
except ConfigNotFound:
    # soundfont_url = "https://archive.org/download/free-soundfonts-sf2-2019-04/Arachno_SoundFont_Version_1.0.sf2" # 148.2 MB !!!
    # acquired from https://musical-artifacts.com/artifacts/2877/CT1MBGMRSV1.06.sf2
    DFLT_SOUNDFONT = str(data_files / "Caeds Small Trash GM v1.06.sf2")

from typing import (
    Callable,
    Iterable,
    List,
    Union,
    Generator,
    Optional,
    Container,
    Any,
    Sequence,
    Tuple,
    Mapping,
    Dict,
)
import os
from operator import attrgetter
import re

from music21.stream import Score, Stream, Part, Measure
from music21.midi.realtime import StreamPlayer
from music21.scale import Scale, MajorScale
from music21.note import Note
from music21 import converter


def identity_func(x):
    return x


def play_music21_object(music21_obj):
    """
    Plays a music21 object (Chord, Part, Stream, etc.) using the StreamPlayer.

    Args:
        music21_obj: A music21 object (Chord, Part, Stream, etc.) to be played.
    """
    if isinstance(music21_obj, Score):
        s = music21_obj
    else:
        # Create a Stream object
        s = Stream()
        # Add the music21 object to the Stream
        s.append(music21_obj)

    # Play the Stream using StreamPlayer
    sp = StreamPlayer(s)
    sp.play()


NoteString = str
NoteSpec = Union[NoteString, Note]
NotesSpec = Iterable[NoteSpec]
TrackSpec = Union[Stream, NotesSpec]
TracksSpec = List[TrackSpec]
ScaleCreator = Callable[[str], Scale]
StreamSpec = Union[str, Iterable[str], List[Note], Stream]
Streams = Iterable[Stream]

StrToNote = Callable[[str], Note]

Filepath = str
ScoreSpec = Union[Filepath, Score]
PartFilter = Callable[[Part], bool]
PartIdx = Union[int, List[int]]
PartFilterSpec = Optional[Union[PartIdx, PartFilter]]

# Regex for note names (e.g., C, C#, Db, etc.)
note_name_pattern = re.compile(r"([A-Ga-g][#b]?)")


def parse_note_name(note_str: str) -> str:
    match = note_name_pattern.match(note_str)
    if match:
        return match.group(1)
    else:
        raise ValueError(f"Invalid note string: {note_str}")


# General utility to add pattern-based aliases to a dictionary
def add_pattern_aliases(quality_extensions):
    for _qe in quality_extensions:
        if _qe.startswith("maj"):
            yield _qe.replace("maj", "M"), quality_extensions[_qe]
        elif _qe.startswith("min"):
            yield _qe.replace("min", "m"), quality_extensions[_qe]
        elif _qe.startswith("dim"):
            yield _qe.replace("dim", "°"), quality_extensions[_qe]


def string_to_note(
    note_or_notes: Optional[Union[NoteString, Iterable[NoteString]]] = None,
    egress: Callable = list,
    **note_kwargs,
) -> Note:
    """
    Converts a string representation of a note to a music21 Note object.

    Args:
        note_or_notes (str or iterable of str): String representation of the note(s).

    Returns:
        Note: The music21 Note object.

    Examples:
        >>> string_to_note('C4')
        <music21.note.Note C>

        If you don't specify a string (only keyword arguments), you get a partial
        function. This is especially useful when you want to convert multiple strings.
        Note also, in the example below, that the function is applied to a list of
        strings.

        >>> my_str_to_note = string_to_note(quarterLength=2, microtone=50)
        >>> notes = list(my_str_to_note(['C4', 'D4', 'E4']))
        >>> note = notes[-1]
        >>> print(f"{note.nameWithOctave=}, {note.duration.type=}, {note.pitch.microtone=}")
        note.nameWithOctave='E4', note.duration.type='half', note.pitch.microtone=<music21.pitch.Microtone (+50c)>

    """
    if note_or_notes is None:
        return partial(string_to_note, **note_kwargs, egress=egress)
    if isinstance(note_or_notes, str):
        return Note(note_or_notes, **note_kwargs)
    elif isinstance(note_or_notes, Iterable):
        return egress(map(partial(string_to_note, **note_kwargs), note_or_notes))


def is_existing_filepath(obj: Any) -> bool:
    """
    Returns True if the input object is a string representing an existing file path.
    """
    return isinstance(obj, str) and os.path.isfile(obj)


def mk_score(obj: ScoreSpec, **kwargs) -> Score:
    """
    Creates a music21 Score object from the input object.

    Args:
        obj (ScoreSpec): The input object.
        **kwargs: Additional keyword arguments to pass to the Score constructor.

    Returns:
        Score: The corresponding Score object.

    >>> score = mk_score([['C4 B4'], ['E4'], ['G4']])
    >>> score.show('text')  # doctest: +ELLIPSIS
    {0.0} <music21.stream.Part 0x...>
        {0.0} <music21.note.Note C>
        {1.0} <music21.note.Note B>
    {2.0} <music21.stream.Part 0x...>
        {0.0} <music21.note.Note E>
    {3.0} <music21.stream.Part 0x...>
        {0.0} <music21.note.Note G>
    """
    if is_existing_filepath(obj):
        return converter.parse(obj)
    if isinstance(obj, list) and obj:
        # assume we have a list of lists, i.e. we have multiple parts
        parts = obj
        score = Score(**kwargs)
        if not parts:  # empty list
            return score
        if not isinstance(obj[0], list):
            parts = [parts]
        for stream_spec in parts:
            part = Part()
            if isinstance(stream_spec, list):
                for note_spec in stream_spec:
                    for note_name in note_spec.split():
                        note = Note(note_name)
                        part.append(note)
            else:
                stream = mk_stream(stream_spec)
                part.append(stream)
            score.append(part)
        return score
    else:
        raise ValueError(f"Cannot create a Score from {obj=}")


def mk_stream(obj: StreamSpec, **kwargs) -> Stream:
    """
    Creates a music21 Stream object from the input object.

    Args:
        obj (StreamSpec): The input object.
        **kwargs: Additional keyword arguments to pass to the Stream constructor.

    Returns:
        Stream: The corresponding Stream object.

    >>> stream = mk_stream('C4 E4 G4')
    >>> note_names(stream)
    ['C4', 'E4', 'G4']
    """
    if isinstance(obj, list) and len(obj) == 1 and isinstance(obj[0], str):
        # if the list has a single string, assume it's a list of note names
        obj = obj[0]  # get the single string of the list (process it in the next step)
    if isinstance(obj, str):
        # first argument is a string, split into list of note names
        obj = obj.split()
    if isinstance(obj, list) and obj:
        if isinstance(obj[0], str):
            # first argument is a list of note names, convert to list of Note objs
            list_of_note_names = obj
            list_of_note_objs = [Note(n) for n in list_of_note_names]
            obj = list_of_note_objs

    return Stream(obj, **kwargs)


def ensure_scale(
    input_scale: Union[str, Scale], scale_creator: ScaleCreator = MajorScale
) -> Scale:
    """
    Ensures the input is a Scale object. Converts a string to a Scale using the provided scale creator.

    Args:
        input_scale (Union[str, Scale]): The input scale.
        scale_creator (Callable[[str], Scale]): A function to create a Scale from a string.

    Returns:
        Scale: The corresponding Scale object.

    Raises:
        AssertionError: If the input cannot be converted to a Scale.
    """
    if isinstance(input_scale, str):
        input_scale = scale_creator(input_scale)
    assert isinstance(
        input_scale, Scale
    ), f"The input scale must be a music21.scale.Scale object: {input_scale=}"
    return input_scale


def get_scale_notes(input_note: Note, input_scale: Scale) -> List[str]:
    """
    Returns the names of the pitches in the scale within one octave around the input note.

    Args:
        input_note (Note): The input note.
        input_scale (Scale): The scale in which to find the pitches.

    Returns:
        List[str]: A list of pitch names within the scale.
    """
    scale_notes = input_scale.getPitches(
        input_note.pitch.transpose(-48), input_note.pitch.transpose(48)
    )
    return [p.nameWithOctave for p in scale_notes]


def ensure_iterable_of_notes(
    notes: TrackSpec, str_to_note: StrToNote = Note
) -> Generator[Note, None, None]:
    """
    Ensures the input is an iterable of Note objects.

    Args:
        notes (TrackSpec): The input notes.

    Returns:
        Generator[Note, None, None]: A generator of Note objects.

    Examples:

    >>> list(ensure_iterable_of_notes(['C4', 'E4', 'G4']))
    [<music21.note.Note C>, <music21.note.Note E>, <music21.note.Note G>]
    >>> s = Stream([Note('C4'), Note('E4'), Note('G4')])
    >>> list(ensure_iterable_of_notes(s))
    [<music21.note.Note C>, <music21.note.Note E>, <music21.note.Note G>]

    """
    if isinstance(notes, Stream):
        return (n for n in notes.notes)

    def _notes():
        for n in notes:
            if isinstance(n, str):
                n = str_to_note(n)
            yield n

    return _notes()


def note_names(notes: TrackSpec, name_attr="nameWithOctave") -> List[str]:
    """
    Returns the names of the notes in the input iterable.

    Args:
        notes (TrackSpec): The input notes.

    Returns:
        List[str]: A list of note names.

    Examples:
        >>> note_names(['C4', 'E4', 'G4'])
        ['C4', 'E4', 'G4']
        >>> s = Stream([Note('C4'), Note('E4'), Note('G4')])
        >>> note_names(s)
        ['C4', 'E4', 'G4']
        >>> score = mk_score([['C4', 'E4', 'G4'], ['A4', 'C5', 'E5']])
        >>> note_names(score)
        [['C4', 'E4', 'G4'], ['A4', 'C5', 'E5']]

    See Also:
        multi_note_names

    """
    if isinstance(notes, Score):
        return list(map(partial(note_names, name_attr=name_attr), notes.parts))
    else:
        return list(map(attrgetter(name_attr), ensure_iterable_of_notes(notes)))


def multi_note_names(tracks: TracksSpec) -> List[List[NoteString]]:
    """
    Returns the names of the notes in the input iterable.

    Args:
        tracks (TracksSpec): The input tracks.

    Returns:
        List[List[NoteString]]: A list of lists of note names.

    Examples:
        >>> tracks = [['C4', 'E4', 'G4'], ['A4', 'C5', 'E5']]
        >>> multi_note_names(tracks)
        [['C4', 'E4', 'G4'], ['A4', 'C5', 'E5']]
        >>> stream1 = Stream([Note('C4'), Note('E4'), Note('G4')])
        >>> stream2 = Stream([Note('A4'), Note('C5'), Note('E5')])
        >>> multi_note_names([stream1, stream2])
        [['C4', 'E4', 'G4'], ['A4', 'C5', 'E5']]
    """
    return [note_names(track) for track in tracks]


def add_streams(list_of_streams: Streams) -> Stream:
    """
    Concatenates multiple streams into a single stream.

    Args:
        list_of_streams (List[Stream]): The list of streams to concatenate.

    Returns:
        Stream: A single stream containing all the concatenated streams.

    Examples:

    >>> s1 = Stream([Note('C4'), Note('E4'), Note('G4')])
    >>> s2 = Stream([Note('A4'), Note('C5'), Note('E5')])
    >>> result_stream = add_streams([s1, s2])
    >>> note_names(result_stream)
    ['C4', 'E4', 'G4', 'A4', 'C5', 'E5']
    """
    combined_stream = Stream()
    for stream in list_of_streams:
        for element in stream:
            combined_stream.append(element)
    return combined_stream


def concatenate_streams(streams_list: List[Streams]) -> Streams:
    """
    Concatenates corresponding streams from a list of lists of streams.

    Args:
        streams_list (List[List[Stream]]): The list of lists of streams to concatenate.

    Returns:
        List[Stream]: A list of concatenated streams.

    Examples:
        >>> s11 = Stream([Note('C4'), Note('E4')])
        >>> s12 = Stream([Note('G4')])
        >>> s21 = Stream([Note('A4')])
        >>> s22 = Stream([Note('C5'), Note('E5')])
        >>> result_streams = concatenate_streams([[s11, s12], [s21, s22]])
        >>> len(result_streams)
        2
        >>> note_names(result_streams[0])
        ['C4', 'E4', 'A4']
        >>> note_names(result_streams[1])
        ['G4', 'C5', 'E5']
    """
    return list(map(add_streams, zip(*streams_list)))


def create_score_from_tracks(tracks: List[Stream]) -> Score:
    """
    Creates a music21 Score from a list of tracks (Stream objects).

    Args:
        tracks (List[Stream]): A list of Stream objects, each representing a track (voice).

    Returns:
        Score: A Score object containing the tracks as separate parts.

    Examples:
        >>> stream1 = Stream([Note('C4'), Note('E4'), Note('G4')])
        >>> stream2 = Stream([Note('A4'), Note('C5'), Note('E5')])
        >>> score = create_score_from_tracks([stream1, stream2])
        >>> len(score.parts)
        2
    """
    parts = [Part(iter(track)) for track in tracks]
    return Score(parts)


def ensure_part_filter(part_filter: PartFilterSpec) -> PartFilter:
    if isinstance(part_filter, int):
        part_filter = [part_filter]
    if isinstance(part_filter, Container):
        # make it into a PartFilter function
        desired_parts = part_filter
        part_filter = lambda i, part: i in desired_parts
    elif part_filter is None:
        part_filter = lambda i, part: True
    elif not callable(part_filter):
        raise TypeError(
            "part_filter must be an int, list of ints, a callable, or None."
            f"Was: {part_filter}"
        )
    return part_filter


def resolve_format_from_filepath(filepath: Filepath) -> str:
    """
    Resolves the format of a file from the file extension.

    Args:
        filepath (Filepath): The file path.

    Returns:
        str: The format of the file.

    Examples:
        >>> resolve_format_from_filepath('output.mid')
        'mid'
        >>> resolve_format_from_filepath('output.wav')
        'wav'
    """
    return os.path.splitext(filepath)[1][1:]


def filter_parts(
    part_filter: PartFilterSpec,
    score_input: ScoreSpec,
    *,
    save_to_filepath: Filepath = None,
) -> Score:
    """
    Filter parts from a score based on the provided filter function or part indices.

    Parameters:
        part_filter: PartFilterSpec
            A function that takes a music21 Part object and returns a boolean value.
            If an integer or list of integers is provided, the function will filter in
            parts based on the indices.
        score_input: Union[str, Score]
            The input score, either as a file path or a music21 Score instance.
        save_to_filepath: str
            The file path to save the modified score. Default is None.

    Returns:
        Score: The score with the subset of parts that were filtered in.

    >>> s = mk_score([['C4'], ['D4'], ['E4']])
    >>> # Test filtering by index
    >>> filtered_score = filter_parts([0, 2], s)
    >>> note_names(filtered_score)
    [['C4'], ['E4']]
    >>> # Test filtering by function
    >>> filtered_score = filter_parts(lambda i, part: part.notes[0].name != 'D', s)
    >>> note_names(filtered_score)
    [['C4'], ['E4']]
    >>> # Test no filtering (return all parts)
    >>> filtered_score = filter_parts(None, s)
    >>> note_names(filtered_score)
    [['C4'], ['D4'], ['E4']]
    """
    filter_func = ensure_part_filter(part_filter)

    # Load the score if a file path is provided
    if isinstance(score_input, str):
        score = converter.parse(score_input)
    else:
        score = score_input

    # Create a new score and add the remaining parts
    new_score = Score()
    for i, part in enumerate(score.parts):
        if filter_func(i, part):
            new_score.append(part)

    # Save the modified score if a file path is provided
    if save_to_filepath:
        # resolve the fmt from the file extension
        fmt = resolve_format_from_filepath(save_to_filepath)
        new_score.write(fmt=fmt, fp=save_to_filepath)

    return new_score


def delete_parts(
    part_idx: PartIdx, score_input: ScoreSpec, *, save_to_filepath: Filepath = None
) -> Score:
    """
    Delete parts from a score based on the provided part indices.

    Parameters:
        part_idx: Union[int, List[int]]
            An integer or a list of integers representing the indices of parts to be deleted.
        score_input: Union[str, Score]
            The input score, either as a file path or a music21 Score instance.
        save_to_filepath: str
            The file path to save the modified score. Default is None.

    Returns:
        Score: The score with specified parts deleted.

    Examples:
        >>> score = mk_score([['C4'], ['D4'], ['E4']])
        >>> modified_score = delete_parts([1], score)
        >>> note_names(modified_score)
        [['C4'], ['E4']]
        >>> modified_score = delete_parts([0, 2], score)
        >>> note_names(modified_score)
        [['D4']]
    """
    if isinstance(part_idx, int):
        part_idx = [part_idx]

    if isinstance(part_idx, str) or not isinstance(part_idx, (Container, Callable)):
        raise ValueError(
            f"part_idx must be an integer or a list of integers. Was: {part_idx}"
        )

    part_filter = lambda i, part: i not in part_idx

    return filter_parts(part_filter, score_input, save_to_filepath=save_to_filepath)



"""Convert music into various formats."""

import os
import io
import tempfile
from typing import Union, Optional, Callable, Any, Dict
from tonal.util import (
    DFLT_MIDI_OUTPUT,
    DFLT_SOUNDFONT,
    DFLT_WAV_OUTPUT,
)

# Type aliases
FilePath = str
BytesData = bytes

dflt_format_for_extension = {
    ".midi": "midi",
    ".mid": "midi",
    ".xml": "musicxml",
    ".musicxml": "musicxml",
    ".wav": "wav",
    ".png": "image",
    ".jpg": "image",
    ".jpeg": "image",
    # '.ipy_audio': 'audio',  # TODO: Handle IPython.display.Audio objects
    # '.ipy_audio': 'ipython',
}

dflt_extension_for_format = {v: k for k, v in dflt_format_for_extension.items()}

# Define magic number signatures for common file formats
FORMAT_SIGNATURES: Dict[bytes, str] = {
    b"MThd": "midi",  # MIDI files start with MThd
    b"RIFF": "wav",  # WAV files start with RIFF
    b"<?xml": "musicxml",  # XML files often start with <?xml
    b"<score-partwise": "musicxml",  # MusicXML specific identifier
    b"\xff\xd8": "image",  # JPEG files start with FF D8
    b"\x89PNG": "image",  # PNG files start with 89 PNG
}


def format_of_filepath(
    filepath: str, *, format_for_extension=dflt_format_for_extension
) -> str:
    """Return the format of a file."""
    ext = os.path.splitext(filepath)[-1].lower()
    return format_for_extension.get(ext, ext)


def replace_extension(src: str, dest_extension: str) -> str:
    """Replace the extension of a filepath with a new one."""
    return src.replace(os.path.splitext(src)[-1], dest_extension)


def ensure_dest_filepath(
    src: str,
    dest: str,
    *,
    dest_format: str = None,
    extension_for_format=dflt_extension_for_format,
) -> str:
    """Ensure that a destination filepath is given."""
    if dest is None:
        src_ext = os.path.splitext(src)[-1]
        dest_ext = extension_for_format.get(dest_format, f".{dest_format}")
        dest = src.replace(src_ext, dest_ext)
    return dest


# Use dol.written_bytes to use file functions as bytes functions
def convert(
    src: Union[FilePath, BytesData],
    dest: Optional[Union[FilePath, str]] = None,
    *,
    src_format: Optional[str] = None,
    dest_format: Optional[str] = None,
    format_for_extension=dflt_format_for_extension,
) -> Union[FilePath, BytesData]:
    """
    Convert between different music formats.

    Args:
        src: Source as a filepath or bytes data
        dest: Destination filepath, format string, or None to return bytes
        src_format: Source format, inferred from filepath extension if None
        dest_format: Destination format, inferred from dest if None
        format_for_extension: Mapping of file extensions to format names

    Returns:
        If dest is a filepath: returns the destination filepath
        If dest is None or a format string or src is bytes: returns the converted bytes

    Examples:
        # File to file conversion
        convert('my_score.musicxml', 'my_score.mid')

        # File to bytes with specified format
        midi_bytes = convert('my_score.musicxml', dest_format='midi')

        # File to bytes using format string shorthand
        midi_bytes = convert('my_score.musicxml', 'midi')

        # Bytes to file
        convert(midi_bytes, 'output.wav', src_format='midi')

        # Bytes to bytes
        wav_bytes = convert(midi_bytes, src_format='midi', dest_format='wav')

        # Bytes to bytes with format string shorthand
        wav_bytes = convert(midi_bytes, 'wav', src_format='midi')
    """
    is_src_bytes = isinstance(src, bytes)

    # If dest is a format string in our known formats, use it as dest_format and set dest to None
    format_values = set(dflt_format_for_extension.values())
    if isinstance(dest, str) and dest in format_values:
        dest_format = dest
        dest = None

    return_bytes = dest is None

    # Determine source format
    if src_format is None:
        if not is_src_bytes:
            # If src is a filepath, infer format from extension
            src_format = format_of_filepath(
                src, format_for_extension=format_for_extension
            )
        else:
            # Try to guess format from bytes magic numbers/headers
            src_format = guess_format_from_bytes(src)

    # If dest is just an extension, add it to the source filename to make the actual filepath
    if not is_src_bytes and isinstance(dest, str) and dest in format_for_extension:
        dest_extension = dest
        dest = replace_extension(src, dest_extension)

    # Determine destination format
    if dest_format is None and dest is not None:
        dest_format = format_of_filepath(
            dest, format_for_extension=format_for_extension
        )

    if not src_format or not dest_format:
        raise ValueError(
            f"Could not determine formats: src_format={src_format}, dest_format={dest_format}. "
            f"Please specify src_format and dest_format explicitly."
        )

    # Handle bytes conversion path
    if is_src_bytes:
        return _convert_from_bytes(
            src, dest, src_format=src_format, dest_format=dest_format
        )

    # Handle file conversion path
    conversion_func = _get_conversion_func(src_format, dest_format)

    if return_bytes:
        # Convert file to bytes
        return _convert_file_to_bytes(src, conversion_func, dest_format)
    else:
        # Regular file to file conversion
        return conversion_func(src, dest)


def guess_format_from_bytes(data: bytes) -> Optional[str]:
    """
    Attempt to guess the format of a file from its bytes content.

    Args:
        data: The bytes data to analyze

    Returns:
        The format string or None if the format couldn't be determined
    """
    if not data:
        return None

    # Check for known file signatures
    for signature, format_name in FORMAT_SIGNATURES.items():
        if data.startswith(signature):
            return format_name

    # Check for MusicXML by looking for common tags if it's likely XML
    if data.startswith(b"<") or b"<?xml" in data[:100]:
        if b"<score-partwise" in data[:1000] or b"<score-timewise" in data[:1000]:
            return "musicxml"

    # Check for MIDI by looking for track chunks if the file starts with MThd
    if len(data) > 14 and data[:4] == b"MThd" and b"MTrk" in data[:1000]:
        return "midi"

    return None


def _get_conversion_func(src_format: str, dest_format: str) -> Callable:
    """Get the appropriate conversion function based on formats."""
    if src_format == "musicxml" and dest_format == "midi":
        return musicxml_to_midi
    elif src_format == "midi" and dest_format == "wav":
        return midi_to_wav
    elif src_format == "image" and dest_format == "musicxml":
        return image_to_musicxml
    else:
        raise ValueError(f"Unsupported conversion: {src_format} -> {dest_format}")


def _convert_from_bytes(
    src_bytes: bytes, dest: Optional[str] = None, *, src_format: str, dest_format: str
) -> Union[FilePath, bytes]:
    """Handle conversions where the source is bytes."""
    # Write bytes to a temporary file with the correct extension
    with tempfile.NamedTemporaryFile(
        suffix=dflt_extension_for_format.get(src_format, f".{src_format}"), delete=False
    ) as temp_file:
        temp_src = temp_file.name
        temp_file.write(src_bytes)

    try:
        conversion_func = _get_conversion_func(src_format, dest_format)

        if dest is None:
            # Return bytes
            temp_dest = temp_src.replace(
                os.path.splitext(temp_src)[1],
                dflt_extension_for_format.get(dest_format, f".{dest_format}"),
            )
            conversion_func(temp_src, temp_dest)

            with open(temp_dest, "rb") as f:
                result = f.read()

            try:
                os.remove(temp_dest)
            except:
                pass

            return result
        else:
            # Write to the specified destination
            return conversion_func(temp_src, dest)
    finally:
        try:
            os.remove(temp_src)
        except:
            pass


def _convert_file_to_bytes(
    src_file: str, conversion_func: Callable, dest_format: str
) -> bytes:
    """Convert a file to bytes using the specified conversion function."""
    # Create a temporary output file
    ext = dflt_extension_for_format.get(dest_format, f".{dest_format}")
    with tempfile.NamedTemporaryFile(suffix=ext, delete=False) as temp_file:
        temp_dest = temp_file.name

    try:
        conversion_func(src_file, temp_dest)
        with open(temp_dest, "rb") as f:
            return f.read()
    finally:
        try:
            os.remove(temp_dest)
        except:
            pass


def image_to_musicxml(
    image: str, *, enable_debug=False, enable_cache=False, remove_teaser_file=True
):
    """Convert and image of a music score into a musicxml file."""
    from homr.main import process_image

    musicxml_filepath, _, teaser_filepath = process_image(
        image, enable_debug=enable_debug, enable_cache=enable_cache
    )

    if remove_teaser_file:
        os.remove(teaser_filepath)

    return musicxml_filepath


def musicxml_to_midi(musicxml_path, midi_path=None):
    """Convert a MusicXML file to MIDI."""
    from music21 import converter

    if midi_path is None:
        midi_path = musicxml_path.replace(".musicxml", ".mid")
    score = converter.parse(musicxml_path)
    score.write("midi", fp=midi_path)
    return midi_path


def midi_to_wav(
    midi_file: str,
    output_wav: str = None,
    *,
    soundfont: str = DFLT_SOUNDFONT,
):
    """
    Synthesize audio from a MIDI file using FluidSynth.

    Args:
        midi_file: Name of the input MIDI file.
        output_wav: Name of the output WAV file.
        soundfont: Path to the SoundFont file.
    """
    import subprocess

    output_wav = ensure_dest_filepath(midi_file, output_wav, dest_format="wav")

    subprocess.run(
        ["fluidsynth", "-ni", soundfont, midi_file, "-F", output_wav, "-r", "44100"]
    )

    return output_wav



"""
Notes, scales, and chords definitions.

A few definitions:

* Scale Quality: The characteristic sound or "flavor" of a scale, determined by its
specific pattern of intervals (semitones) from the root note.
Examples include major, minor, pentatonic, blues, or whole-tone.
* Chord Quality: The characteristic sound or "flavor" of a chord, determined by its
specific combination of intervals (semitones) above its root note.
Common examples include major, minor, diminished, augmented, or dominant 7th.
* Semitone Pattern: A numerical representation of a musical scale or chord, showing
the precise distance in semitones (half-steps) of each note from the starting (root)
note. For example, the semitone pattern for a major scale is [0, 2, 4, 5, 7, 9, 11].

"""

from typing import Dict, Sequence, Tuple
from tonal.util import (
    note_name_pattern,
    parse_note_name,
    add_pattern_aliases,
)

import re


# Define root notes to MIDI note numbers
root_notes_anglo: Dict[str, int] = {
    # Anglo notation
    "C": 60,
    "C#": 61,
    "Db": 61,
    "D": 62,
    "D#": 63,
    "Eb": 63,
    "E": 64,
    "F": 65,
    "F#": 66,
    "Gb": 66,
    "G": 67,
    "G#": 68,
    "Ab": 68,
    "A": 69,
    "A#": 70,
    "Bb": 70,
    "B": 71,
}
root_notes_latin_to_anglo = {
    "do": "C",
    "do#": "C#",
    "reb": "C#",
    "re": "D",
    "re#": "D#",
    "mib": "D#",
    "mi": "E",
    "fa": "F",
    "fa#": "F#",
    "solb": "F#",
    "sol": "G",
    "sol#": "G#",
    "lab": "G#",
    "la": "A",
    "la#": "A#",
    "sib": "A#",
    "si": "B",
}

# Make a name-to-midinote mapping with root_notes_anglo and root_notes_latin_to_anglo
root_notes = dict(
    **root_notes_anglo,
    **{k: root_notes_anglo[v] for k, v in root_notes_latin_to_anglo.items()},
)

# Regex to match any root at the start, case-insensitive, preferring the longest key.
# This allows patterns like "Bbmajor" (no separator) or "F#_minor".
_root_keys_longest_first = sorted(root_notes.keys(), key=len, reverse=True)
_root_alt = "|".join(re.escape(k) for k in _root_keys_longest_first)
root_note_re = re.compile(rf"^({_root_alt})", re.IGNORECASE)


def _rebuild_root_note_re() -> None:
    """Rebuild the root note regex from the current root_notes keys.

    Must be called whenever root_notes is modified dynamically (e.g., via register_root_note).
    """
    global _root_keys_longest_first, _root_alt, root_note_re
    _root_keys_longest_first = sorted(root_notes.keys(), key=len, reverse=True)
    _root_alt = "|".join(re.escape(k) for k in _root_keys_longest_first)
    root_note_re = re.compile(rf"^({_root_alt})", re.IGNORECASE)


scale_quality = {
    # Western Diatonic & Common Scales (retained as before)
    "major": (0, 2, 4, 5, 7, 9, 11),
    "natural_minor": (0, 2, 3, 5, 7, 8, 10),
    "harmonic_minor": (0, 2, 3, 5, 7, 8, 11),
    "melodic_minor_ascending": (0, 2, 3, 5, 7, 9, 11),
    "melodic_minor_descending": (0, 2, 3, 5, 7, 8, 10),
    "major_pentatonic": (0, 2, 4, 7, 9),
    "minor_pentatonic": (0, 3, 5, 7, 10),
    "blues": (0, 3, 5, 6, 7, 10),
    "chromatic": (0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11),
    "whole_tone": (0, 2, 4, 6, 8, 10),
    "diminished": (0, 2, 3, 5, 6, 8, 9, 11),
    "diminished_half_whole": (0, 1, 3, 4, 6, 7, 9, 10),
    "augmented": (0, 3, 4, 7, 8, 11),
    "lydian": (0, 2, 4, 6, 7, 9, 11),
    "mixolydian": (0, 2, 4, 5, 7, 9, 10),
    "dorian": (0, 2, 3, 5, 7, 9, 10),
    "phrygian": (0, 1, 3, 5, 7, 8, 10),
    "locrian": (0, 1, 3, 4, 6, 8, 10),
    "aeolian": (0, 2, 3, 5, 7, 8, 10),
    "ionian": (0, 2, 4, 5, 7, 9, 11),
    # Major Jazz Scales (distinct or commonly named in jazz)
    "bebop_dominant": (
        0,
        2,
        4,
        5,
        7,
        9,
        10,
        11,
    ),  # Mixolydian with added M7 (or b7, M7)
    "bebop_major": (0, 2, 4, 5, 7, 8, 9, 11),  # Major with added b6 (or M6, b6)
    "bebop_minor": (0, 2, 3, 4, 5, 7, 9, 10),  # Dorian with added M3 (or m3, M3)
    "bebop_melodic_minor": (0, 2, 3, 5, 7, 8, 9, 11),  # Melodic Minor with added b6
    "altered": (0, 1, 3, 4, 6, 8, 10),  # 7th mode of Melodic Minor, aka Super Locrian
    "lydian_dominant": (
        0,
        2,
        4,
        6,
        7,
        9,
        10,
    ),  # 4th mode of Melodic Minor, aka Mixolydian #11
    "phrygian_dominant": (
        0,
        1,
        4,
        5,
        7,
        8,
        10,
    ),  # 5th mode of Harmonic Minor (also common in Middle Eastern music)
    # Indian Ragas (distinct from Western scales, or now canonical here)
    # Re-checking semitone patterns and prioritizing jazz names
    "todi": (0, 1, 3, 6, 7, 8, 11),  # Hindustani Todi Thaat
    "poorvi": (0, 1, 4, 6, 7, 8, 11),  # Hindustani Poorvi Thaat
    "marwa": (0, 1, 4, 6, 7, 9, 11),  # Hindustani Marwa Thaat
    "chalanata": (0, 3, 5, 6, 7, 10, 11),  # Melakarta 36
    "hamsadhwani": (0, 4, 7, 9, 11),  # Pentatonic raga
    # Potential clashes:
    "bhairav": (0, 1, 4, 5, 7, 8, 11),  # Hindustani Bhairav Thaat
    "chakravakam": (0, 1, 4, 5, 7, 9, 10),  # Melakarta 16
    "malkauns": (0, 3, 5, 8, 10),  # Pentatonic raga
}

_scale_quality_western_aliases = {
    # Western Aliases
    "": "major",
    "maj": "major",
    "M": "major",
    "min": "natural_minor",
    "m": "natural_minor",
    "minor": "natural_minor",
    "harmonic": "harmonic_minor",
    "melodic": "melodic_minor_ascending",
    "penta_major": "major_pentatonic",
    "penta_min": "minor_pentatonic",
    "pentatonic": "major_pentatonic",
    "penta": "major_pentatonic",
    "minor_penta": "minor_pentatonic",
    "major_penta": "major_pentatonic",
    "whl_tone": "whole_tone",
    "diminished": "diminished",
    "dim": "diminished",
    "diminished (whole-half)": "diminished",
    "diminished (half-whole)": "diminished_half_whole",
    "aug": "augmented",
    "chrom": "chromatic",
    "blues_scale": "blues",
    "dor": "dorian",
    "phr": "phrygian",
    "lyd": "lydian",
    "mix": "mixolydian",
    "aeo": "aeolian",
    "ion": "ionian",
    "nat_min": "natural_minor",
    "harm_min": "harmonic_minor",
    "mel_min_asc": "melodic_minor_ascending",
    "mel_min_desc": "melodic_minor_descending",
    "w-h_dim": "diminished",
    "h-w_dim": "diminished_half_whole",
}

_scale_quality_jazz_aliases = {
    # Jazz Aliases
    "bebop_dom": "bebop_dominant",
    "bebop_maj": "bebop_major",
    "bebop_min": "bebop_minor",
    "bebop_dorian": "bebop_minor",  # Common alias for bebop minor
    "bebop_melodic": "bebop_melodic_minor",
    "alt": "altered",
    "super_locrian": "altered",
    "lyd_dom": "lydian_dominant",
    "mixolydian_#11": "lydian_dominant",
    "acoustic_scale": "lydian_dominant",  # Also sometimes called this
    "phryg_dom": "phrygian_dominant",
    "dominant_phrygian": "phrygian_dominant",
    "mixolydian_b2_b6": "phrygian_dominant",  # If you want to get descriptive
}

_scale_quality_raga_aliases = {
    # Indian Raga Aliases
    # Aliases for Ragas whose patterns match Western/Jazz scales:
    "bilawal": "major",  # Hindustani Thaat for Major
    "ionian_raga": "major",
    "shankarabharanam": "major",  # Carnatic Melakarta 29 for Major
    "dheerasankarabharanam": "major",  # Carnatic Melakarta 29 for Major
    "kharaharapriya": "dorian",  # Carnatic Melakarta 22 for Dorian
    "melakarta_22": "dorian",
    "kafi": "dorian",  # Hindustani Thaat for Dorian
    "harikambhoji": "mixolydian",  # Carnatic Melakarta 28 for Mixolydian
    "melakarta_28": "mixolydian",
    "khamaj": "mixolydian",  # Hindustani Thaat for Mixolydian
    "natabhairavi": "natural_minor",  # Carnatic Melakarta 20 for Natural Minor
    "melakarta_20": "natural_minor",
    "asavari": "natural_minor",  # Hindustani Thaat for Natural Minor / Aeolian
    "aeolian_raga": "natural_minor",
    "hanumatodi": "phrygian",  # Carnatic Melakarta 8 for Phrygian
    "melakarta_8": "phrygian",
    "phrygian_raga": "phrygian",
    "kalyan": "lydian",  # Hindustani Thaat for Lydian
    "kirwani": "harmonic_minor",  # Melakarta 21, identical to Harmonic Minor
    "melakarta_21": "harmonic_minor",
    "mayamalavagowla": "bhairav",  # Carnatic Melakarta 15
    "melakarta_15": "bhairav",
    "bhupali": "major_pentatonic",  # Hindustani Raga identical to Major Pentatonic
    "mohnam": "major_pentatonic",  # Carnatic equivalent of Bhupali / Major Pentatonic
    "durga": "major_pentatonic",
    "deshkar": "major_pentatonic",
    "carnatic_major": "major",  # Common reference for the Carnatic major scale
    "carnatic_minor": "natural_minor",  # Common reference for the Carnatic natural minor scale
    # Aliases for Ragas unique to scale_quality (or now canonical here)
    "bhairav_thaat": "bhairav",
    "todi_thaat": "todi",
    "poorvi_thaat": "poorvi",
    "marwa_thaat": "marwa",
    "melakarta_36": "chalanata",
    "melakarta_16": "chakravakam",
    "hansadhwani_raga": "hamsadhwani",
    "malkauns_raga": "malkauns",
    "indian_m_minor": "malkauns",  # A less common, but sometimes used, description
}

_scale_quality_backcompat_aliases = {
    # Back-compat aliases mapping spaced names to underscored canonical names
    "natural_minor": "natural_minor",
    "harmonic_minor": "harmonic_minor",
    "melodic_minor_ascending": "melodic_minor_ascending",
    "melodic_minor_descending": "melodic_minor_descending",
    "major_pentatonic": "major_pentatonic",
    "minor_pentatonic": "minor_pentatonic",
    "whole_tone": "whole_tone",
    "bebop_dominant": "bebop_dominant",
    "bebop_major": "bebop_major",
    "bebop_minor": "bebop_minor",
    "bebop_melodic_minor": "bebop_melodic_minor",
    "lydian_dominant": "lydian_dominant",
    "phrygian_dominant": "phrygian_dominant",
}

scale_quality_alias = {
    **_scale_quality_western_aliases,
    **_scale_quality_jazz_aliases,
    **_scale_quality_raga_aliases,
    **_scale_quality_backcompat_aliases,
}

# assert that all values of scale_quality_alias are in scale_quality
assert all(
    v in scale_quality for v in scale_quality_alias.values()
), "Some scale quality aliases are not in the main scale quality definitions."

# TODO: Separate base and aliases
# TODO: Verify completeness and more chord definitions if needed
# TODO: See if defs can be infered from parsing the chord names
# Define quality and extension intervals
chord_quality: Dict[str, Sequence[int]] = {
    "": (0, 4, 7),  # Major triad, 'C' -> 'Cmaj
    "maj": (0, 4, 7),  # Major triad
    "M": (0, 4, 7),  # Major triad
    "m": (0, 3, 7),  # Minor triad
    "min": (0, 3, 7),  # Minor triad
    "dim": (0, 3, 6),  # Diminished triad
    "aug": (0, 4, 8),  # Augmented triad
    "7": (0, 4, 7, 10),  # Dominant 7th
    "7M": (0, 4, 7, 11),  # Major 7th
    "maj7": (0, 4, 7, 11),  # Major 7th
    "min7": (0, 3, 7, 10),  # Minor 7th
    "minmaj7": (0, 3, 7, 11),  # Minor major 7th
    "dim7": (0, 3, 6, 9),  # Diminished 7th
    "hdim7": (0, 3, 6, 10),  # Half-diminished 7th
    "aug7": (0, 4, 8, 10),  # Augmented 7th
    "6": (0, 4, 7, 9),  # Major 6th
    "min6": (0, 3, 7, 9),  # Minor 6th
    "9": (0, 4, 7, 10, 14),  # Dominant 9th
    "maj9": (0, 4, 7, 11, 14),  # Major 9th
    "min9": (0, 3, 7, 10, 14),  # Minor 9th
    "11": (0, 4, 7, 10, 14, 17),  # Dominant 11th
    "maj11": (0, 4, 7, 11, 14, 17),  # Major 11th
    "min11": (0, 3, 7, 10, 14, 17),  # Minor 11th
    "13": (0, 4, 7, 10, 14, 17, 21),  # Dominant 13th
    "maj13": (0, 4, 7, 11, 14, 17, 21),  # Major 13th
    "min13": (0, 3, 7, 10, 14, 17, 21),  # Minor 13th
}

# -----------------------------------------------------------------------------
# Registration
# -----------------------------------------------------------------------------


# TODO: Need to update global root_note_re value, since root_notes changes here:
def register_root_note(note_name: str, midi_value: int) -> None:
    """Register a new root note with its MIDI value.

    Args:
        note_name: The name of the note (e.g., 'C', 'C#', 'Db')
        midi_value: The MIDI note number (0-127)

    >>> register_root_note('H', 71)  # German notation for B
    >>> root_notes['H']
    71
    """
    if not (0 <= midi_value <= 127):
        raise ValueError(f"MIDI value must be between 0 and 127, got {midi_value}")
    root_notes[note_name] = midi_value
    # Update regex to recognize the newly added root at parse time
    _rebuild_root_note_re()


def register_scale_quality(quality_name: str, semitone_pattern: Sequence[int]) -> None:
    """Register a new scale quality with its semitone pattern.

    Args:
        quality_name: The name of the scale quality
        semitone_pattern: Sequence of semitone intervals from root

    >>> register_scale_quality('custom_scale', (0, 2, 5, 7, 10))
    >>> scale_quality['custom_scale']
    (0, 2, 5, 7, 10)
    """
    pattern_tuple = tuple(semitone_pattern)
    if not all(isinstance(x, int) and 0 <= x <= 11 for x in pattern_tuple):
        raise ValueError("Semitone pattern must contain integers between 0 and 11")
    if 0 not in pattern_tuple:
        raise ValueError("Semitone pattern must include 0 (the root note)")
    scale_quality[quality_name] = pattern_tuple


def register_scale_quality_alias(alias: str, canonical_name: str) -> None:
    """Register an alias for an existing scale quality.

    Args:
        alias: The alias name
        canonical_name: The existing scale quality name or alias

    >>> register_scale_quality_alias('my_major', 'major')
    >>> scale_quality_alias['my_major']
    'major'
    """
    # Verify that the canonical name exists (either in scale_quality or as an alias)
    try:
        semitone_pattern(canonical_name)
    except ValueError:
        raise ValueError(
            f"Canonical name '{canonical_name}' is not a valid scale quality"
        )
    scale_quality_alias[alias] = canonical_name


def register_chord_quality(quality_name: str, semitone_pattern: Sequence[int]) -> None:
    """Register a new chord quality with its semitone pattern.

    Args:
        quality_name: The name of the chord quality
        semitone_pattern: Sequence of semitone intervals from root

    >>> register_chord_quality('my_chord', (0, 4, 7, 10, 13))
    >>> chord_quality['my_chord']
    (0, 4, 7, 10, 13)
    """
    pattern_tuple = tuple(semitone_pattern)
    if not all(isinstance(x, int) and x >= 0 for x in pattern_tuple):
        raise ValueError("Semitone pattern must contain non-negative integers")
    if 0 not in pattern_tuple:
        raise ValueError("Semitone pattern must include 0 (the root note)")
    chord_quality[quality_name] = pattern_tuple


def list_root_notes() -> Dict[str, int]:
    """List all registered root notes.

    Returns:
        Dictionary mapping note names to MIDI values

    >>> notes = list_root_notes()
    >>> 'C' in notes
    True
    >>> notes['C']
    60
    """
    return dict(root_notes)


def list_scale_qualities(include_aliases: bool = False) -> Dict[str, Sequence[int]]:
    """List all registered scale qualities.

    Args:
        include_aliases: If True, include aliases in the listing

    Returns:
        Dictionary mapping quality names to semitone patterns

    >>> qualities = list_scale_qualities()
    >>> 'major' in qualities
    True
    >>> qualities['major']
    (0, 2, 4, 5, 7, 9, 11)
    """
    result = dict(scale_quality)
    if include_aliases:
        for alias, canonical in scale_quality_alias.items():
            if canonical in scale_quality:
                result[alias] = scale_quality[canonical]
            elif canonical in scale_quality_alias:
                # Handle alias chains
                try:
                    pattern = semitone_pattern(canonical)
                    result[alias] = pattern
                except ValueError:
                    pass  # Skip invalid aliases
    return result


def list_scale_quality_aliases() -> Dict[str, str]:
    """List all registered scale quality aliases.

    Returns:
        Dictionary mapping aliases to canonical names

    >>> aliases = list_scale_quality_aliases()
    >>> 'maj' in aliases
    True
    >>> aliases['maj']
    'major'
    """
    return dict(scale_quality_alias)


def list_chord_qualities() -> Dict[str, Sequence[int]]:
    """List all registered chord qualities.

    Returns:
        Dictionary mapping quality names to semitone patterns

    >>> qualities = list_chord_qualities()
    >>> 'maj' in qualities
    True
    >>> qualities['maj']
    (0, 4, 7)
    """
    return dict(chord_quality)


# TODO: Make a framework for user-defined aliases


# -----------------------------------------------------------------------------
# Validation & Utilities
# -----------------------------------------------------------------------------


# Define a specific error for incorrect scale specifications
class IncorrectScaleSpecification(ValueError):
    """Raised when a scale string cannot be parsed/validated into root and quality."""


def list_scales_string() -> str:
    """Return a user-friendly help string describing scale syntax and valid values.

    Includes:
    - Anatomy of a scale specification
    - Valid roots
    - Valid qualities (including aliases)
    """
    valid_roots = sorted(list_root_notes().keys())
    valid_roots_fmt = ", ".join(f"'{r}'" for r in valid_roots)

    valid_qualities = sorted(list_scale_qualities(include_aliases=True).keys())
    # show qualities as a wrapped, indented list for readability
    qualities_fmt = "\n  - " + "\n  - ".join(
        [q if q else "<empty string>" for q in valid_qualities]
    )

    return (
        "Scale specification anatomy: A root and quality separated by an underscore, "
        "a space, or nothing (so, '<root>_<quality>', '<root> <quality>' "
        "or '<root><quality>') (case-insensitive).\n"
        "- Root is optional (defaults to 'C' in some contexts).\n"
        "- Empty quality means 'major' (alias '').\n\n"
        f"Valid roots: {valid_roots_fmt}.\n"
        f"Valid qualities (including aliases):{qualities_fmt}\n"
    )


def validate_scale_semitone_pattern_uniquness(scale_quality: dict) -> bool:
    """
    Validates that all scale qualities have unique semitone patterns.
    """
    seen = {}
    for name, pattern in scale_quality.items():
        pattern_tuple = tuple(pattern)
        if pattern_tuple in seen:
            raise ValueError(
                f"Scale quality '{name}' has a non-unique semitone pattern. "
                f"Duplicates: {seen[pattern_tuple]} and {name}"
            )
        seen[pattern_tuple] = name
    return True


def validate_scale_aliases(scale_quality: dict, scale_aliases: dict) -> bool:
    """
    Validates that all values in the scale_aliases dictionary are valid keys
    in the scale_qualities dictionary.

    Args:
        scale_quality (dict): The dictionary of canonical scale qualities and patterns.
        scale_aliases (dict): The dictionary of scale aliases.

    Returns:
        bool: True if all aliases are valid, False otherwise.
    """
    all_valid = True
    invalid_aliases = []
    for alias_key, canonical_name in scale_aliases.items():
        if canonical_name not in scale_quality:
            print(
                f"Validation Error: Alias '{alias_key}' points to "
                f"non-existent scale quality '{canonical_name}' in scale_quality."
            )
            invalid_aliases.append(alias_key)
            all_valid = False

    if not all_valid:
        raise ValueError(f"Invalid aliases found: {invalid_aliases}")


# validate_scale_semitone_pattern_uniquness(scale_quality)  # Disabled: duplicate patterns are allowed for now  # TODO: move duplicates to aliases
validate_scale_aliases(scale_quality, scale_quality_alias)


# -----------------------------------------------------------------------------
# Access and Usage
# -----------------------------------------------------------------------------


def semitone_pattern(quality: str) -> tuple:
    """
    Get the semitone pattern for a given scale quality string.
    Looks in scale_quality, then in scale_quality_alias (using the alias to look in scale_quality).
    Raises ValueError if not found.

    >>> semitone_pattern('major')
    (0, 2, 4, 5, 7, 9, 11)
    >>> semitone_pattern('maj')
    (0, 2, 4, 5, 7, 9, 11)
    >>> semitone_pattern('')
    (0, 2, 4, 5, 7, 9, 11)
    >>> semitone_pattern('dorian')
    (0, 2, 3, 5, 7, 9, 10)

    """
    if quality in scale_quality:
        return scale_quality[quality]
    elif quality in scale_quality_alias:
        alias = scale_quality_alias[quality]
        if alias in scale_quality:
            return scale_quality[alias]
        elif alias in scale_quality_alias:
            # Handle alias chaining (e.g., '' -> 'major' -> scale_quality)
            return semitone_pattern(alias)
    raise ValueError(f"Unknown scale quality: {quality}")


def canonical_quality_name(quality: str) -> str:
    """Return the canonical quality key from `scale_quality`.

    Accepts aliases (including chains) and normalizes spaces to underscores.
    Raises ValueError if no mapping exists.

    Examples:
        - "minor" -> "natural_minor"
        - "bebop_dom" -> "bebop_dominant"
        - "minor pentatonic" -> "minor_pentatonic"
    """
    raw = (quality or "").strip().lower()
    # Generate a set of reasonable variants to try against aliases and canonical keys
    variants = []

    def add(v: str):
        if v not in variants:
            variants.append(v)

    add(raw)
    add(raw.replace(" ", "_"))
    add(raw.replace("_", " "))
    no_paren = re.sub(r"[()]", "", raw)
    add(no_paren)
    add(no_paren.replace(" ", "_"))
    add(no_paren.replace("_", " "))

    for q in variants:
        if q in scale_quality:
            return q
        # chase aliases starting from this variant
        seen = set()
        cur = q
        while cur in scale_quality_alias and cur not in seen:
            seen.add(cur)
            cur = scale_quality_alias[cur]
            if cur in scale_quality:
                return cur
    # Not found
    raise ValueError(f"Unknown scale quality: {quality}")


def scale_params(scale: str, midi_notes: bool = False):
    """
    Parse a scale specification string and return (root_note, scale_quality).
    If midi_notes is True, returns (root_note_midi, semitone_pattern).
    If midi_notes is False, returns (root_note_str, scale_quality_str).
    The root_note can be an empty string (meaning default root).

    >>> scale_params('C major')
    ('C', 'major')
    >>> scale_params('dorian')
    ('', 'dorian')
    >>> scale_params('C')
    ('C', 'major')
    >>> scale_params('C', midi_notes=True)
    (60, (0, 2, 4, 5, 7, 9, 11))
    >>> scale_params('dorian', midi_notes=True)
    (None, (0, 2, 3, 5, 7, 9, 10))

    """
    s = scale.strip()
    m = root_note_re.match(s)
    root = ""
    quality_raw = s.lower()
    if m:
        root_raw = m.group(0)
        candidate_root = next(
            (k for k in root_notes if k.lower() == root_raw.lower()), root_raw
        )
        remaining = s[len(root_raw) :]
        candidate_quality = remaining.lstrip(" _").lower()
        # If only root is provided, use alias ''
        if candidate_quality == "":
            root = candidate_root
            quality_raw = ""
        else:
            # Only accept the matched root if the remaining is a valid quality
            try:
                canonical_quality_name(candidate_quality)
            except ValueError:
                # Treat whole string as a quality (e.g., 'dorian' shouldn't split as 'do'+'rian')
                root = ""
                quality_raw = s.lower()
            else:
                root = candidate_root
                quality_raw = candidate_quality

    try:
        canonical_quality = canonical_quality_name(quality_raw)
        pattern = semitone_pattern(canonical_quality)
    except ValueError as e:
        raise IncorrectScaleSpecification(
            (
                f"Incorrect scale specification: '{scale}' (unknown scale quality '{quality_raw}').\n"
                + list_scales_string()
            )
        ) from e

    # Normalize root to Anglo spelling if needed
    if root:
        # ensure we use the canonical Anglo key when a Latin root is provided
        lower_root = root.lower()
        if lower_root in root_notes_latin_to_anglo:
            root = root_notes_latin_to_anglo[lower_root]
        # ensure the case matches a key in root_notes_anglo if already anglo
        elif root not in root_notes_anglo:
            # try find matching key in anglo ignoring case
            for k in root_notes_anglo:
                if k.lower() == lower_root:
                    root = k
                    break

    if midi_notes:
        midi = None
        if root:
            midi = next(
                (v for k, v in root_notes.items() if k.lower() == root.lower()), None
            )
        return midi, pattern
    else:
        # return canonical names
        return root, canonical_quality


def scale_midi_notes(
    scale: str = "C major",
    midi_range: Tuple[int, int] = (0, 127),
    *,
    default_root: str = "C",
) -> tuple:
    """
    Return a tuple of all MIDI note numbers in the given range that belong to the specified scale.

    Args:
        scale (str): The scale string, e.g., 'C major', 'D# minor pentatonic'.
        midi_range (tuple): The (min, max) MIDI note numbers to include.
        default_root (str): The root note to use if not found in the scale string (default 'C').

    Returns:
        tuple: MIDI note numbers in the scale within the specified range.


    >>> scale_midi_notes('E')  # doctest: +ELLIPSIS
    (1, 3, 4, 6, 8, 9, 11, 13, 15, 16, ...114, 116, 117, 119, 121, 123, 124, 126)

    If it's suprising that it starts at 1, because you were expecting 'E' (major)
    to start with 4, remember that scale_midi_notes is designed to return all notes in the scale
    within the MIDI range, not just the notes starting from the root note.
    The first note in that range that is in E major is actually a C#, which the
    midi note 1 is. You can control the range through:

    >>> scale_midi_notes('E', midi_range=(4, 30))  # doctest: +ELLIPSIS
    (4, 6, 8, 9, 11, 13, 15, 16, 18, 20, 21, 23, 25, 27, 28, 30)

    You see that if no scale quality is specified, it defaults to 'major'.
    On the other hand, if no root note is specified, it defaults to 'C', or what ever
    you tell `default_root` is should be.

    >>> assert (
    ...     scale_midi_notes('', midi_range=(60, 72))
    ...     == scale_midi_notes('C', midi_range=(60, 72))
    ...     == scale_midi_notes('major', midi_range=(60, 72))
    ...     == scale_midi_notes('C major', midi_range=(60, 72))
    ...     == (60, 62, 64, 65, 67, 69, 71, 72)
    ... )

    >>> scale_midi_notes('Db minor pentatonic', midi_range=(60, 72))
    (61, 64, 66, 68, 71)

    """
    # Parse root and pattern
    root, quality = scale_params(scale)
    if not root:
        root = default_root
    if root not in root_notes:
        raise ValueError(f"Unknown root note: {root}")
    root_midi = root_notes[root]
    pattern = semitone_pattern(quality)
    notes = []
    # Find the first root note in the midi_range
    first_root = root_midi
    while first_root < midi_range[0]:
        first_root += 12
    while first_root > midi_range[0]:
        first_root -= 12
    # Generate all scale notes in the range
    for midi_note in range(first_root, midi_range[1] + 1, 12):
        for interval in pattern:
            n = midi_note + interval
            if midi_range[0] <= n <= midi_range[1]:
                notes.append(n)
    return tuple(sorted(set(notes)))



"""Tools for chords and chord progressions."""

from mido import MidiFile, MidiTrack, Message
import subprocess
from typing import Sequence, Tuple, Callable, Mapping, Dict
import os
import re
from config2py import process_path
from tonal.util import (
    DFLT_OUTPUT_NAME,
    DFLT_MIDI_OUTPUT,
    DFLT_SOUNDFONT,
    note_name_pattern,
    parse_note_name,
    add_pattern_aliases,
)
from tonal.notes import root_notes, chord_quality
from music21.note import Note

DFLT_CHORD_SEQUENCE = [
    ("Bdim", 120),
    ("Em11", 120),
    ("Amin9", 120),
    ("Dm7", 120),
    "G7",
    "Cmaj7",
]

# Type aliases for this module
Chord = str
Notes = Sequence[Note]
ChordTimed = Tuple[Chord, Notes]
ChordSequence = Sequence[ChordTimed]
ChordDefinitions = Callable[[Chord], Notes]
ChordRenderer = Callable[[Notes, any, int], None]


def chord_to_notes(chord: Chord) -> Notes:
    """
    Parse a chord string and return the corresponding sequence of MIDI note numbers.

    :param chord: The chord string (e.g., 'Cmaj7').
    :return: A sequence of MIDI note numbers representing the chord.
    """
    root = parse_note_name(chord)
    quality_extension = chord[len(root) :]
    root_midi = root_notes.get(root)
    # print(root, root_midi, quality_extension)

    if root_midi is None:
        raise ValueError(f"Unknown root note: {root}")

    intervals = chord_quality.get(quality_extension)

    if intervals is None:
        raise ValueError(f"Unknown quality/extension: {quality_extension}")

    return [root_midi + interval for interval in intervals]


if not os.path.exists(DFLT_SOUNDFONT):
    from warnings import warn

    warn(f"Soundfont not found at {DFLT_SOUNDFONT}")


chord_renders = {}


def register_chord_render(chord_renderer: ChordRenderer, name=None):
    if name is None:
        name = chord_renderer.__name__
    chord_renders[chord_renderer.__name__] = chord_renderer
    return chord_renderer


@register_chord_render
def play_simultaneously(
    notes: Sequence[int], track: MidiTrack, duration: int, *, velocity=64
):
    v = velocity
    for note in notes:
        track.append(Message("note_on", note=note, velocity=v, time=0))
    _notes = iter(notes)
    track.append(Message("note_off", note=next(_notes), velocity=v, time=duration))
    for note in _notes:
        track.append(Message("note_off", note=note, velocity=v, time=0))


# TODO: Need a better implementation for arpeggios.
@register_chord_render
def play_arpeggio(
    notes: Sequence[int], track: MidiTrack, duration: int, *, velocity=64
):
    v = velocity
    note_duration = duration // len(notes)
    current_time = 0
    for note in notes:
        track.append(Message("note_on", note=note, velocity=v, time=current_time))
        current_time += note_duration
        track.append(Message("note_off", note=note, velocity=v, time=current_time))


def resolve_chord_render(chord_renderer: ChordRenderer) -> ChordRenderer:
    if isinstance(chord_renderer, str):
        name = chord_renderer
        chord_renderer = chord_renders.get(name)
        if chord_renderer is None:
            raise ValueError(
                f"Unknown chord renderer: {name}. "
                '(Available: {", ".join(chord_renders)}). '
                "You can register a new chord renderer with `register_chord_render`."
            )
        return chord_renderer
    return chord_renderer


DFLT_DURATION = 240 * 4


def process_chord_sequence(
    chord_sequence: ChordSequence, default_duration=DFLT_DURATION
):
    """Preprocess a chord sequence, to make sure to add time, etc."""
    for chord in chord_sequence:
        if isinstance(chord, str):
            yield chord, default_duration
        elif isinstance(chord, tuple):
            yield chord
        else:
            raise ValueError(f"Invalid chord: {chord}")


def chords_to_midi(
    chord_sequence: ChordSequence = DFLT_CHORD_SEQUENCE,
    *,
    output_file: str = None,  # DFLT_MIDI_OUTPUT,
    render_chord: ChordRenderer = play_simultaneously,
    chord_definitions: ChordDefinitions = chord_to_notes,
):
    """
    Generate a MIDI file from a chord sequence.

    :param chord_sequence: List of tuples containing chords and their duration.
    :param chord_definitions: Dictionary mapping chords to MIDI note patterns.
    :param output_file: Name of the output MIDI file.
    :param render_chord: Function defining how the chords should be played.
    """
    midi = MidiFile()
    track = MidiTrack()
    midi.tracks.append(track)

    render_chord = resolve_chord_render(render_chord)
    chord_sequence = process_chord_sequence(chord_sequence)

    track.append(Message("program_change", program=0, time=0))  # Acoustic Grand Piano

    for chord, duration in chord_sequence:
        pattern = chord_definitions(chord)
        if pattern:
            render_chord(pattern, track, duration)

    if output_file:
        midi.save(output_file)
        return output_file
    else:
        return midi


def chords_to_wav(
    chord_sequence: ChordSequence = DFLT_CHORD_SEQUENCE,
    name: str = DFLT_OUTPUT_NAME,
    *,
    chord_definitions: ChordDefinitions = chord_to_notes,
    soundfont: str = DFLT_SOUNDFONT,
    render_chord: ChordRenderer = play_simultaneously,
):
    """
    Generate a WAV file directly from a chord sequence.

    :param chord_sequence: List of tuples containing chords and their duration.
    :param name: Base name for the output MIDI and WAV files.
    :param chord_definitions: Dictionary mapping chords to MIDI note patterns.
    :param soundfont: Path to the SoundFont file.
    :param render_chord: Function defining how the chords should be played.
    """
    from sonification.converters import midi_to_wav

    midi_file = f"{name}.mid"
    wav_file = f"{name}.wav"

    chords_to_midi(
        chord_sequence,
        chord_definitions=chord_definitions,
        output_file=midi_file,
        render_chord=render_chord,
    )
    midi_to_wav(midi_file=midi_file, output_wav=wav_file, soundfont=soundfont)
    return wav_file



"""Test notes module"""

import pytest

from tonal.notes import (
    list_root_notes,
    list_scale_qualities,
    scale_midi_notes,
    scale_params,
    list_scales_string,
    IncorrectScaleSpecification,
)


def test_list_scales_string_has_expected_sections():
    s = list_scales_string()
    sl = s.lower()
    assert "anatomy" in sl
    assert "valid roots" in sl
    assert "valid qualities" in sl


def test_all_root_quality_combinations_parse_with_scale_midi_notes():
    roots = sorted(list_root_notes().keys())
    qualities = sorted(list_scale_qualities(include_aliases=True).keys())

    for root in roots:
        for quality in qualities:
            spec = f"{root} {quality}".strip()
            try:
                notes = scale_midi_notes(spec)
            except IncorrectScaleSpecification as e:
                pytest.fail(f"Spec '{spec}' raised IncorrectScaleSpecification: {e}")
            # Should be a tuple of ints (can be empty depending on range, but default range is full MIDI)
            assert isinstance(notes, tuple)
            assert all(isinstance(n, int) for n in notes)


def test_scale_params_raises_helpful_error_on_invalid_spec():
    with pytest.raises(IncorrectScaleSpecification) as excinfo:
        scale_params("H hypermix")
    msg = str(excinfo.value).lower()
    assert "incorrect scale specification" in msg
    assert "valid roots" in msg
    assert "valid qualities" in msg


def test_scale_params_returns_normalized_root_and_quality():
    """
    scale_params should normalize:
    - Root names to Anglo notation (using latin->anglo map when needed)
    - Quality names to canonical keys from `scale_quality` (resolving aliases)
    Also verify flexibility of separators:
    - Space vs underscore in the quality
    - Space, underscore, or nothing between root and quality
    """
    assert scale_params("C# bebop dominant") == ("C#", "bebop_dominant")
    assert scale_params("C# bebop_dom") == ("C#", "bebop_dominant")
    assert scale_params("do#_bebop dominant") == ("C#", "bebop_dominant")
    assert scale_params("do#bebop_dominant") == ("C#", "bebop_dominant")
    # more normalization cases
    assert scale_params("la minor") == ("A", "natural_minor")
    assert scale_params("sib major penta") == ("A#", "major_pentatonic")
    assert scale_params("Bb_major_pentatonic") == ("Bb", "major_pentatonic")
    assert scale_params("Db minor pentatonic") == ("Db", "minor_pentatonic")



"""
Tools for counterpoint.

The `translate_in_scale` allows you to translate a sequence of notes, or multiple
tracks of notes by the given number of steps within the given scale.

>>> stream = translate_in_scale(['C4', 'E4', 'B3', 'C4'], -2, 'C')
>>> stream  # doctest: +ELLIPSIS
<music21.stream.Stream ...>
>>> note_names(stream)
['A3', 'C4', 'G3', 'A3']

For multiple tracks:

>>> motif = [['C4', 'E4', 'G4'], ['A4', 'C5', 'E5']]
>>> translated_tracks = translate_in_scale(motif, -2, 'C')
>>> multi_note_names(translated_tracks)
[['A3', 'C4', 'E4'], ['F4', 'A4', 'C5']]

Using some other scales:

With a E major scale:

>>> motif = [['E4', 'G#4', 'B4'], ['C#5', 'E5', 'G#5']]
>>> translated_tracks = translate_in_scale(motif, 1, 'E')
>>> multi_note_names(translated_tracks)
[['F#4', 'A4', 'C#5'], ['D#5', 'F#5', 'A5']]

With a D flat major scale:

>>> motif = [['Db4', 'F4', 'Ab4'], ['Bb4', 'Db5', 'F5']]
>>> translated_tracks = translate_in_scale(motif, -3, 'Db')
>>> multi_note_names(translated_tracks)
[['A-3', 'C4', 'E-4'], ['F4', 'A-4', 'C5']]

Now let's use a different, "custom" scale, as well as demonstrate the use
of a partial function to get a translator with a fixed input scale:

>>> from functools import partial
>>> from music21.scale import HarmonicMinorScale
>>> translate = partial(
...     translate_in_scale, input_scale='A', scale_creator=HarmonicMinorScale
... )
>>> motif = [['A4', 'C5', 'E5'], ['G#5', 'A5', 'C6']]
>>> translated_tracks = translate(motif, 2)
>>> multi_note_names(translated_tracks)
[['C5', 'E5', 'G#5'], ['B5', 'C6', 'E6']]
"""

from typing import List, Union, Iterable
from music21.scale import Scale, MajorScale
from music21.stream import Stream
from music21.note import Note
from tonal.util import (
    NoteString,
    TrackSpec,
    TracksSpec,
    ScaleCreator,
    note_names,
    multi_note_names,
    get_scale_notes,
    ensure_scale,
    ensure_iterable_of_notes,
    concatenate_streams,
)


def _translate_note_in_scale(
    input_note: NoteString,
    translation: int,
    input_scale: Union[str, Scale],
    *,
    scale_creator: ScaleCreator = MajorScale,
) -> str:
    """
    Translates the input note by the given number of steps within the given scale.

    Args:
        input_note (str): The input note (e.g., 'C4', 'D#5').
        translation (int): The number of steps to translate the note.
        input_scale (Union[str, Scale]): The scale in which to perform the translation.
        scale_creator (ScaleCreator): A function to create a Scale from a string.

    Returns:
        str: The translated note.

    Raises:
        ValueError: If the input note is not in the specified scale.

    Examples:
        >>> _translate_note_in_scale('C4', 0, 'C')
        'C4'
        >>> _translate_note_in_scale('E4', -2, 'C')
        'C4'
        >>> _translate_note_in_scale('B4', 3, 'C')
        'E5'

        Use the E major scale
        >>> _translate_note_in_scale('E4', 1, 'E')
        'F#4'
        >>> _translate_note_in_scale('G#4', -1, 'E')
        'F#4'
        >>> _translate_note_in_scale('B4', 2, 'E')
        'D#5'

        Use the D flat major scale
        >>> _translate_note_in_scale('Db4', -1, 'Db')
        'C4'
        >>> _translate_note_in_scale('F4', 2, 'Db')
        'A-4'
        >>> _translate_note_in_scale('Ab4', -3, 'Db')
        'E-4'

        Now let's use a different, "custom" scale, as well as demonstrate the use
        of a partial function to get a translator with a fixed input scale:

        >>> from functools import partial
        >>> from music21.scale import HarmonicMinorScale
        >>> translate = partial(
        ...     _translate_note_in_scale, input_scale='A', scale_creator=HarmonicMinorScale
        ... )
        >>> translate('A4', 2)
        'C5'
        >>> translate('C5', -2)
        'A4'
        >>> translate('C5', 4)
        'G#5'
        >>> translate('G#5', 1)
        'A5'
    """
    n = Note(input_note)
    input_scale = ensure_scale(input_scale, scale_creator)
    scale_pitch_names = get_scale_notes(n, input_scale)

    # Find the index of the input note in the scale
    try:
        index = scale_pitch_names.index(n.nameWithOctave)
    except ValueError:
        raise ValueError("The input note is not in the specified scale.")

    # Calculate the new index with wrap around
    new_index = (index + translation) % len(scale_pitch_names)

    return scale_pitch_names[new_index]


def translate_notes_in_scale(
    input_notes: TrackSpec,
    translation: int,
    input_scale: Union[str, Scale],
    *,
    scale_creator: ScaleCreator = MajorScale,
) -> Stream:
    """
    Translates a sequence of notes by the given number of steps within the given scale.

    Args:
        input_notes (TrackSpec): The input notes.
        translation (int): The number of steps to translate the notes.
        input_scale (Union[str, Scale]): The scale in which to perform the translation.
        scale_creator (ScaleCreator): A function to create a Scale from a string.

    Returns:
        Stream: A stream of translated notes.

    Examples:

    >>> result_stream = translate_notes_in_scale(['C4', 'E4', 'B3', 'C4'], -2, 'C')
    >>> note_names(result_stream)
    ['A3', 'C4', 'G3', 'A3']
    """
    input_notes = ensure_iterable_of_notes(input_notes)
    return Stream(
        [
            Note(
                _translate_note_in_scale(
                    n.nameWithOctave,
                    translation,
                    input_scale,
                    scale_creator=scale_creator,
                )
            )
            for n in input_notes
        ]
    )


def translate_in_scale(
    motif: Union[TrackSpec, TracksSpec],
    translation: Union[int, Iterable[int]],
    input_scale: Union[str, Scale],
    *,
    scale_creator: ScaleCreator = MajorScale,
) -> Union[Stream, List[Stream]]:
    """
    Translates the input notes or tracks by the given number of steps within the given scale.

    Args:
        motif (Union[TrackSpec, TracksSpec]): The motif; input notes or tracks.
        translation (int): The number of steps to translate the notes.
        input_scale (Union[str, Scale]): The scale in which to perform the translation.
        scale_creator (ScaleCreator): A function to create a Scale from a string.

    Returns:
        Union[Stream, List[Stream]]: A stream of translated notes or a list of streams of translated notes.

    Examples:
        For a single track of notes:

        >>> stream = translate_in_scale(['C4', 'E4', 'B3', 'C4'], -2, 'C')
        >>> stream  # doctest: +ELLIPSIS
        <music21.stream.Stream ...>
        >>> note_names(stream)
        ['A3', 'C4', 'G3', 'A3']

        For multiple tracks:

        >>> motif = [['C4', 'E4', 'G4'], ['A4', 'C5', 'E5']]
        >>> translated_tracks = translate_in_scale(motif, -2, 'C')
        >>> multi_note_names(translated_tracks)
        [['A3', 'C4', 'E4'], ['F4', 'A4', 'C5']]

        Using some other scales:

        With a E major scale, and two translations:

        >>> motif = [['E4', 'G#4', 'B4'], ['C#5', 'E5', 'G#5']]
        >>> translated_tracks = translate_in_scale(motif, [1, 2], 'E')
        >>> multi_note_names(translated_tracks)  # doctest: +NORMALIZE_WHITESPACE
        [['F#4', 'A4', 'C#5', 'G#4', 'B4', 'D#5'],
        ['D#5', 'F#5', 'A5', 'E5', 'G#5', 'B5']]

        With a D flat major scale:

        >>> motif = [['Db4', 'F4', 'Ab4'], ['Bb4', 'Db5', 'F5']]
        >>> translated_tracks = translate_in_scale(motif, -3, 'Db')
        >>> multi_note_names(translated_tracks)
        [['A-3', 'C4', 'E-4'], ['F4', 'A-4', 'C5']]

        Now let's use a different, "custom" scale, as well as demonstrate the use
        of a partial function to get a translator with a fixed input scale:

        >>> from functools import partial
        >>> from music21.scale import HarmonicMinorScale
        >>> translate = partial(
        ...     translate_in_scale, input_scale='A', scale_creator=HarmonicMinorScale
        ... )
        >>> motif = [['A4', 'C5', 'E5'], ['G#5', 'A5', 'C6']]
        >>> translated_tracks = translate(motif, 2)
        >>> multi_note_names(translated_tracks)
        [['C5', 'E5', 'G#5'], ['B5', 'C6', 'E6']]
    """
    if isinstance(translation, Iterable):
        tracks_to_concatinate = [
            translate_in_scale(motif, t, input_scale, scale_creator=scale_creator)
            for t in translation
        ]
        return concatenate_streams(tracks_to_concatinate)

    else:
        if isinstance(motif, list) and all(
            isinstance(item, (list, Stream)) for item in motif
        ):
            translated_tracks = [
                translate_notes_in_scale(
                    track, translation, input_scale, scale_creator=scale_creator
                )
                for track in motif
            ]
            return translated_tracks
        else:
            if isinstance(motif, (str, Note)):
                motif = [motif]  # in case it's just a single note
            return translate_notes_in_scale(
                motif, translation, input_scale, scale_creator=scale_creator
            )
